from cupyx import jit


@jit.rawkernel(device="cuda")
def get_this_agent_data_from_tensor(agent_index, property_tensor):
    return property_tensor[agent_index]


@jit.rawkernel(device="cuda")
def set_this_agent_data_from_tensor(agent_index, property_tensor, value):
    property_tensor[agent_index] = value


@jit.rawkernel(device="cuda")
def get_neighbor_data_from_tensor(agent_ids, neighbor_id, property_tensor):
    neighbor_index = -1
    i = 0
    while i < len(agent_ids) and int(agent_ids[i]) != int(neighbor_id):
        i += 1
    if i < len(agent_ids):
        neighbor_index = i
    return property_tensor[neighbor_index]


@jit.rawkernel(device="cuda")
def set_neighbor_ids_for_network_space(agent_ids, neighbor_id, property_tensor, value):
    neighbor_index = -1
    i = 0
    while i < len(agent_ids) and int(agent_ids[i]) != int(neighbor_id):
        i += 1
    if i < len(agent_ids):
        neighbor_index = i
    property_tensor[neighbor_index] = value



from sagesim.model import Model
from sagesim.space import NetworkSpace
from sir_breed import SIRBreed


class SIRModel(Model):
    """
    SIRModel class for the SIR model.
    Inherits from the Model class in the sagesim library.
    """

    def __init__(self, p_infection=0.2, p_recovery=0.2) -> None:
        space = NetworkSpace()
        super().__init__(space)
        self._sir_breed = SIRBreed()

        # Register the breed
        self.register_breed(breed=self._sir_breed)

        # register user-defined global properties
        self.register_global_property("p_infection", p_infection)
        self.register_global_property("p_recovery", p_recovery)

    # create_agent method takes user-defined properties, that is, the state and preventative_measures, to create an agent
    def create_agent(self, state, preventative_measures):
        agent_id = self.create_agent_of_breed(
            self._sir_breed, state=state, preventative_measures=preventative_measures
        )
        return agent_id

    def connect_agents(self, agent_0, agent_1):
        self.get_space().connect_agents(agent_0, agent_1)



from time import time
import argparse
from sir_model import SIRModel
from state import SIRState
from mpi4py import MPI
from random import random

from random import sample

import networkx as nx

comm = MPI.COMM_WORLD
num_workers = comm.Get_size()
worker = comm.Get_rank()


def generate_small_world_network(n, k, p):
    """
    Generate a small world network using the Watts-Strogatz model.

    Parameters:
    - n (int): The number of nodes in the network.
    - k (int): Each node is connected to its k nearest neighbors in a ring topology.
    - p (float): The probability of rewiring each edge.

    Returns:
    - networkx.Graph: The generated small world network.
    """
    return nx.watts_strogatz_graph(n, k, p)


def generate_small_world_of_agents(
    model, num_agents: int, num_init_connections: int, num_infected: int
) -> SIRModel:
    network = generate_small_world_network(num_agents, num_init_connections, 0.2)
    for n in network.nodes:
        preventative_measures = [random() for _ in range(100)]
        model.create_agent(SIRState.SUSCEPTIBLE.value, preventative_measures)

    """print(f"Number of infected agents: {num_infected}")"""
    for n in sample(sorted(network.nodes), num_infected):
        model.set_agent_property_value(n, "state", SIRState.INFECTED.value)

    for edge in network.edges:
        model.connect_agents(edge[0], edge[1])
    return model


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--num_agents",
        type=int,
    )
    parser.add_argument(
        "--num_init_connections",
        type=float,
    )
    parser.add_argument(
        "--num_nodes",
        type=int,
    )
    args = parser.parse_args()

    model = SIRModel()
    model.setup(use_gpu=True)
    num_agents = args.num_agents
    num_init_connections = int(args.num_init_connections)
    num_nodes = args.num_nodes

    model_creation_start = time()
    model = generate_small_world_of_agents(
        model, num_agents, num_init_connections, int(0.1 * num_agents)
    )  # test_network()  #
    model_creation_end = time()
    model_creation_duration = model_creation_end - model_creation_start
    """print(
        [
            SIRState(model.get_agent_property_value(agent_id, property_name="state"))
            for agent_id in range(n_agents)
        ]
    )"""

    simulate_start = time()
    model.simulate(10, sync_workers_every_n_ticks=1)
    simulate_end = time()
    simulate_duration = simulate_end - simulate_start

    if worker == 0:
        with open("execution_times.csv", "a") as f:
            f.write(
                f"{num_agents}, {num_init_connections}, {num_nodes}, {num_workers}, {model_creation_duration}, {simulate_duration}\n"
            )

    result = [
        SIRState(model.get_agent_property_value(agent_id, property_name="state"))
        for agent_id in range(num_agents)
        if model.get_agent_property_value(agent_id, property_name="state") is not None
    ]

    """if worker == 0:
        print(
            [
                SIRState(
                    model.get_agent_property_value(agent_id, property_name="state")
                )
                for agent_id in range(num_agents)
            ]
        )

        print(result)"""



from time import time
import argparse
from typing import OrderedDict
from sfr_model import SFRModel
from mpi4py import MPI
import random
from random import sample
import json
import pandas as pd

import networkx as nx

comm = MPI.COMM_WORLD
num_workers = comm.Get_size()
worker = comm.Get_rank()


def generate_small_world_of_agents(model, network) -> SFRModel:
    # create agent with random popularity value and init number of vehicles
    for i, n in enumerate(network.nodes):
        model.create_agent(
            # osmnxid=network.nodes[n]["osmnxid"],
            popularity=random.random(),
            vehicle_num=random.randint(0, 10),
        )

    for edge in network.edges:
        if (
            edge[0] < model._agent_factory.num_agents
            and edge[1] < model._agent_factory.num_agents
        ):
            model.connect_agents(edge[0], edge[1])
    return model


if __name__ == "__main__":

    # load the sf drivng network
    # load json file
    with open(f"sfr.json", "r") as f:
        G_info = json.load(f)

    nodes = G_info["nodes"]
    edges = G_info["edges"]

    # create a muliiDiGraph using the node and edge lists
    G = nx.MultiDiGraph()

    # List of nodes
    G.add_nodes_from(nodes)

    # Add edges with attributes
    for u, v, k, length, maxspeed in edges:
        G.add_edge(u, v, key=k, length=length, maxspeed=maxspeed)

    # create a identical network with integer node id
    # Create mapping from original node ID to integers
    original_nodes = list(nodes)
    osmnx2idx = OrderedDict((node, idx) for idx, node in enumerate(original_nodes))
    osmnxids = list(osmnx2idx.keys())

    # Create a new MultiDiGraph with masked node IDs
    # [TO-do] Maybe a easier way to re-key the nodes
    G_masked = nx.MultiDiGraph()
    G_masked.add_nodes_from(range(len(original_nodes)))

    for osmnxid, masked_id in osmnx2idx.items():
        G_masked.add_node(masked_id, osmnxid=osmnxid)

    # Add edges using the new node IDs and preserve attributes
    for u, v, k, data in G.edges(keys=True, data=True):
        new_u = osmnx2idx[u]
        new_v = osmnx2idx[v]
        G_masked.add_edge(new_u, new_v, key=k, **data)

    model = SFRModel()
    model.setup(use_gpu=True)
    model = generate_small_world_of_agents(model, G_masked)

    simulate_start = time()
    model.simulate(10, sync_workers_every_n_ticks=1)
    simulate_end = time()
    simulate_duration = simulate_end - simulate_start

    if worker == 0:
        model_creation_start = time()
        model_creation_duration = model_creation_start - simulate_start
        print(f"Model creation took {model_creation_duration} seconds")
        print(f"Simulation took {simulate_duration} seconds")

    checked = set()
    result = []
    for agent_idx in range(len(nodes)):
        vn = model.get_agent_property_value(agent_idx, property_name="vehicle_num")
        osmnxid = osmnxids[agent_idx]
        if osmnxid in checked:
            print(f"Duplicate osmnxid found: {osmnxid}")
            exit()
        checked.add(osmnxid)
        result.append(
            {
                "osmnxid": osmnxid,
                "popularity": model.get_agent_property_value(
                    agent_idx, property_name="popularity"
                ),
                "vehicle_num": vn,
            }
        )
    df = pd.DataFrame(result)
    df.to_csv("sfr_results.csv", index=False)

    # if worker == 0:
    #     with open("execution_times.csv", "a") as f:
    #         f.write(
    #             f"{num_agents}, {num_edges}, {num_nodes}, {num_workers}, {model_creation_duration}, {simulate_duration}\n"
    #         )

    # if worker == 0:
    #     print(
    #         [
    #             SFRState(
    #                 model.get_agent_property_value(agent_id, property_name="state")
    #             )
    #             for agent_id in range(num_agents)
    #         ]
    #     )



from enum import Enum
from random import random

from sagesim.breed import Breed

from state import SIRState
from sir_step_func import step_func


# Define the SIRState enumeration for agent states
class SIRState(Enum):
    SUSCEPTIBLE = 1
    INFECTED = 2
    RECOVERED = 3


class SIRBreed(Breed):
    """
    SIRBreed class the SIR model.
    Inherits from the Breed class in the sagesim library.
    """

    def __init__(self) -> None:
        name = "SIR"
        super().__init__(name)
        # Register properties for the breed
        self.register_property("state", SIRState.SUSCEPTIBLE.value)
        self.register_property("preventative_measures", [random() for _ in range(100)])
        # Register the step function
        self.register_step_func(step_func, "sir_step_func.py", 0)



from __future__ import annotations
from typing import Any, Callable, Iterable, List, Dict
from collections import OrderedDict
from copy import copy
import sys

import numpy as np
from mpi4py import MPI

from sagesim.breed import Breed
from sagesim.internal_utils import (
    compress_tensor,
)
from sagesim.space import Space


comm = MPI.COMM_WORLD
num_workers = comm.Get_size()
worker = comm.Get_rank()


class AgentFactory:
    def __init__(self, space: Space) -> None:
        self._breeds: Dict[str, Breed] = OrderedDict()
        self._space: Space = space
        self._space._agent_factory = self
        self._num_breeds = 0
        self._num_agents = 0
        self._property_name_2_agent_data_tensor = OrderedDict(
            {"breed": [], "locations": []}
        )
        self._property_name_2_defaults = OrderedDict(
            {
                "breed": 0,
                "locations": self._space._locations_defaults,
            }
        )
        self._property_name_2_index = {
            "breed": 0,
            "locations": 1,
        }
        self._agent2rank = {}  # global
        self._rank2agentid2agentidx = {}  # global

        self._current_rank = 0

        self._prev_agent_data = {}

    @property
    def breeds(self) -> List[Breed]:
        """
        Returns the breeds registered in the model

        :return: A list of currently registered breeds.

        """
        return self._breeds.values()

    @property
    def num_agents(self) -> int:
        """
        Returns number of agents. Agents are not removed if they are killed at the
            moment.

        """
        return self._num_agents

    @property
    def num_properties(self) -> int:
        """
        Returns number of properties, equivalent to the number
        of agent data tensors.

        """
        return len(self._property_name_2_agent_data_tensor)

    def register_breed(self, breed: Breed) -> None:
        """
        Registered agent breed in the model so that agents can be created under
            this definition.

        :param breed: Breed definition of agent

        """
        breed._breedidx = self._num_breeds
        self._num_breeds += 1
        self._breeds[breed.name] = breed
        for property_name, default in breed.properties.items():
            if property_name in self._property_name_2_agent_data_tensor:
                # If the property is already registered, just update the default value
                self._property_name_2_defaults[property_name] = default
            else:
                # Register the new property
                self._property_name_2_index[property_name] = len(
                    self._property_name_2_agent_data_tensor
                )
                self._property_name_2_agent_data_tensor[property_name] = []
                self._property_name_2_defaults[property_name] = default

    def create_agent(self, breed: Breed, **kwargs) -> int:
        """
        Creates and agent of the given breed initialized with the properties given in
            **kwargs.

        :param breed: Breed definition of agent
        :param **kwargs: named arguments of agent properties. Names much match properties
            already registered in breed.
        :return: Agent ID

        """

        agent_id = self._num_agents
        # Assign agents to rank in round robin fashion across available workers.
        self._agent2rank[agent_id] = self._current_rank
        agentid2agentidx_of_current_rank = self._rank2agentid2agentidx.get(
            self._current_rank, OrderedDict()
        )
        agentid2agentidx_of_current_rank[agent_id] = len(
            self._property_name_2_agent_data_tensor["locations"]
        )
        self._rank2agentid2agentidx[self._current_rank] = (
            agentid2agentidx_of_current_rank
        )

        self._current_rank += 1
        if self._current_rank >= num_workers:
            self._current_rank = 0

        if worker == self._agent2rank[agent_id]:
            # Only the worker that owns this agent will create and store the agent data.
            # This is to avoid unnecessary data duplication across workers.
            if breed.name not in self._breeds:
                raise ValueError(f"Fatal: unregistered breed {breed.name}")
            property_names = self._property_name_2_agent_data_tensor.keys()
            for property_name in property_names:
                if property_name == "breed":
                    breed = self._breeds[breed.name]
                    self._property_name_2_agent_data_tensor[property_name].append(
                        breed._breedidx
                    )
                else:
                    default_value = copy(self._property_name_2_defaults[property_name])
                    self._property_name_2_agent_data_tensor[property_name].append(
                        kwargs.get(property_name, default_value)
                    )

        self._num_agents += 1

        return agent_id

    def get_agent_property_value(self, property_name: str, agent_id: int) -> Any:
        """
        Returns the value of the specified property_name of the agent with
            agent_id

        :param property_name: str name of property as registered in the breed.
        :param agent_id: Agent's id as returned by create_agent
        :return: value of property_name property for agent of agent_id
        """
        agent_rank = self._agent2rank[agent_id]
        if agent_rank == worker:
            subcontextidx = self._rank2agentid2agentidx.get(worker).get(agent_id)
            result = self._property_name_2_agent_data_tensor[property_name][
                subcontextidx
            ]
        else:
            result = None
        result = comm.bcast(result, root=agent_rank)

        return result

    def set_agent_property_value(
        self,
        property_name: str,
        agent_id: int,
        value: Any,
    ) -> None:
        """
        Sets the property of property_name for the agent with agent_id with
            value.
        :param property_name: str name of property as registered in the breed.
        :param agent_id: Agent's id as returned by create_agent
        :param value: New value for property
        """
        if worker == self._agent2rank[agent_id]:
            if property_name not in self._property_name_2_agent_data_tensor:
                raise ValueError(f"{property_name} not a property of any breed")
            subcontextidx = self._rank2agentid2agentidx.get(worker).get(agent_id)
            self._property_name_2_agent_data_tensor[property_name][
                subcontextidx
            ] = value

    def get_agents_with(self, query: Callable) -> Dict[int, List[Any]]:
        """
        Returns an Dict, key: agent_id value: List of properties, of the agents that satisfy
            the query. Query must be a callable that returns a boolean and accepts **kwargs
            where arguments may with breed property names may be accepted and used to form
            query logic.

        :param query: Callable that takes agent data as dict and returns List of agent data
        :return: Dict of agent_id: List of properties

        """
        matching_agents = {}
        property_names = self._property_name_2_agent_data_tensor.keys()
        for agent_id in range(self._num_agents):
            agent_properties = {
                property_name: self._property_name_2_agent_data_tensor[property_name][
                    agent_id
                ]
                for property_name in property_names
            }
            if query(**agent_properties):
                matching_agents[agent_id] = agent_properties
        return matching_agents

    def _generate_agent_data_tensors(
        self,
    ) -> List[List[Any]]:
        """converted_agent_data_tensors = []
        for property_name in self._property_name_2_agent_data_tensor.keys():
            converted_agent_data_tensors.append(
                convert_to_equal_side_tensor(
                    self._property_name_2_agent_data_tensor[property_name]
                )
            )

        return converted_agent_data_tensors"""
        return list(self._property_name_2_agent_data_tensor.values())

    def _update_agent_property(
        self,
        regularized_agent_data_tensors,
        agent_id: int,
        property_name: str,
    ) -> None:
        if worker == self._agent2rank[agent_id]:
            subcontextidx = self._rank2agentid2agentidx.get(worker).get(agent_id)
            property_idx = self._property_name_2_index[property_name]
            adt = regularized_agent_data_tensors[property_idx]
            value = (
                compress_tensor(adt[subcontextidx], min_axis=0)
                if type(adt[subcontextidx]) == Iterable
                else adt[subcontextidx]
            )

            self._property_name_2_agent_data_tensor[property_name][
                subcontextidx
            ] = value

    def contextualize_agent_data_tensors(
        self, agent_data_tensors, agent_ids_chunk, all_neighbors
    ):
        """
        Chunks agent data tensors so that each distributed worker does not
        get more data than the agents that worker processes actually need.

        :return: 2-tuple.
            1. agent_ids_chunks: List of Lists of agent_ids to be processed
                by each worker.
            3. agent_data_tensors_subcontexts: subcontext of agent_data_tensors
                required by agents of agent_ids_chunks to be processed by a worker
        """

        neighborrank2agentidandadt = {}
        neighborrankandagentidsvisited = set()
        num_agents_this_rank = len(agent_ids_chunk)
        for agent_idx in range(num_agents_this_rank):
            agent_id = agent_ids_chunk[agent_idx]
            agent_adts = [adt[agent_idx] for adt in agent_data_tensors]
            if agent_id not in self._prev_agent_data:
                self._prev_agent_data[agent_id] = agent_adts
            else:
                # If the agent data has not changed, skip sending it
                agent_changed = False
                for prop_idx in range(self.num_properties):
                    if not np.array_equal(
                        agent_adts[prop_idx],
                        self._prev_agent_data[agent_id][prop_idx],
                        equal_nan=True,
                    ):
                        agent_changed = True
                        break
                if agent_changed:
                    # Update the previous agent data
                    self._prev_agent_data[agent_id] = agent_adts
                else:
                    # Skip sending this agent if its data has not changed
                    continue

            for neighbor_id in all_neighbors[agent_idx]:
                if np.isnan(neighbor_id):
                    break
                neighbor_rank = self._agent2rank[int(neighbor_id)]
                if neighbor_rank == worker:
                    # Don't send to self
                    continue
                if (neighbor_rank, agent_id) not in neighborrankandagentidsvisited:
                    # Don't send the same agent to the same rank multiple times
                    neighborrankandagentidsvisited.add((neighbor_rank, agent_id))
                    if neighbor_rank not in neighborrank2agentidandadt.keys():
                        neighborrank2agentidandadt[neighbor_rank] = []
                    neighborrank2agentidandadt[neighbor_rank].append(
                        (agent_id, agent_adts)
                    )

        received_neighbor_adts = []
        received_neighbor_ids = []
        # Send chunk nums
        sends_num_chunks = []
        torank2numchunks = {}
        total_num_chunks = 0
        other_ranks_to = [(worker + i) % num_workers for i in range(1, num_workers)]
        other_ranks_from = [(worker + i) % num_workers for i in range(1, num_workers)]
        # Calculate chunk_size to ensure each chunk is <= 128 bytes
        # Estimate the size of a single value in neighborrank2agentidandadt
        if neighborrank2agentidandadt:
            sample_value = next(iter(neighborrank2agentidandadt.values()))[0]
            estimated_value_size = sys.getsizeof(
                sample_value
            )  # Approximate size in bytes
            chunk_size = max(
                1, 128 // estimated_value_size
            )  # Ensure at least one value per chunk
        else:
            chunk_size = 0  # Default to 1 if no data is present
        for to_rank in other_ranks_to:
            if to_rank in neighborrank2agentidandadt:
                # Send the data for this rank
                data_to_send_to_rank = neighborrank2agentidandadt[to_rank]
                # Break the data into chunks

                num_chunks = len(data_to_send_to_rank) // chunk_size + (
                    1 if len(data_to_send_to_rank) % chunk_size > 0 else 0
                )
                total_num_chunks += num_chunks
                torank2numchunks[to_rank] = num_chunks
                sends_num_chunks.append(
                    comm.isend(
                        num_chunks,
                        dest=to_rank,
                        tag=0,
                    )
                )
            else:
                # No data to send to this rank
                torank2numchunks[to_rank] = 0
                sends_num_chunks.append(
                    comm.isend(
                        0,
                        dest=to_rank,
                        tag=0,
                    )
                )
        # Receive num_chunks from all ranks
        recvs_num_chunks_requests = []
        for from_rank in other_ranks_from:
            recvs_num_chunks_requests.append(comm.irecv(source=from_rank, tag=0))

        MPI.Request.waitall(sends_num_chunks)
        recvs_num_chunks = MPI.Request.waitall(recvs_num_chunks_requests)

        # Send the chunks
        send_chunk_requests = []
        for to_rank in other_ranks_to:
            if to_rank in neighborrank2agentidandadt:
                # Send the data for this rank
                data_to_send_to_rank = neighborrank2agentidandadt[to_rank]
                num_chunks = torank2numchunks[to_rank]
                for i in range(num_chunks):
                    chunk = data_to_send_to_rank[i * chunk_size : (i + 1) * chunk_size]
                    send_chunk_request = comm.isend(
                        chunk,
                        dest=to_rank,
                        tag=i + 1,
                    )
                    if i >= len(send_chunk_requests):
                        send_chunk_requests.append([])
                    send_chunk_requests[i].append(send_chunk_request)
        # Receive the chunks
        recv_chunk_requests = []
        for i, from_rank in enumerate(other_ranks_from):
            num_chunks = recvs_num_chunks[i]
            for j in range(num_chunks):
                received_chunk_request = comm.irecv(source=from_rank, tag=j + 1)
                if j >= len(recv_chunk_requests):
                    recv_chunk_requests.append([])
                recv_chunk_requests[j].append(received_chunk_request)

        received_data = []
        num_send_chunk_requests = len(send_chunk_requests)
        num_recv_chunk_requests = len(recv_chunk_requests)
        for i in range(max(num_send_chunk_requests, num_recv_chunk_requests)):
            if i < num_send_chunk_requests:
                MPI.Request.waitall(send_chunk_requests[i])
            if i < num_recv_chunk_requests:
                received_data_ranks_chunk = MPI.Request.waitall(recv_chunk_requests[i])
                for received_data_rank_chunk in received_data_ranks_chunk:
                    received_data.extend(received_data_rank_chunk)

        # Process received chunks
        received_neighbor_adts = [[] for _ in range(self.num_properties)]
        received_neighbor_ids = []
        for neighbor_idx, (neighbor_id, adts) in enumerate(received_data):
            received_neighbor_ids.append(neighbor_id)
            for prop_idx in range(self.num_properties):
                received_neighbor_adts[prop_idx].append(adts[prop_idx])

        return (
            agent_ids_chunk,
            agent_data_tensors,
            received_neighbor_ids,
            received_neighbor_adts,
        )

    def reduce_agent_data_tensors(
        self,
        agent_and_neighbor_data_tensors,
        agent_and_neighbor_ids_in_subcontext,
        reduce_func: Callable = None,
    ):

        num_agents_this_rank = len(self._rank2agentid2agentidx.get(worker).keys())
        agent_ids = agent_and_neighbor_ids_in_subcontext[:num_agents_this_rank]
        neighbor_ids = agent_and_neighbor_ids_in_subcontext[num_agents_this_rank:]

        agent_data_tensors = [
            agent_and_neighbor_data_tensors[prop_idx][:num_agents_this_rank].tolist()
            for prop_idx in range(self.num_properties)
        ]
        neighbor_adts = [
            agent_and_neighbor_data_tensors[i][num_agents_this_rank:].tolist()
            for i in range(self.num_properties)
        ]

        # Find rank of neighbors
        neighbors_visited = set()
        neighborrank2neighboridandadt = OrderedDict()
        for neighbor_idx, neighbor_id in enumerate(neighbor_ids):
            if neighbor_id not in neighbors_visited:
                if np.isnan(neighbor_id):
                    continue
                neighbors_visited.add(neighbor_id)
                neighbor_rank = self._agent2rank[neighbor_id]
                neighbor_adt = [
                    neighbor_adt[neighbor_idx] for neighbor_adt in neighbor_adts
                ]
                if neighbor_rank not in neighborrank2neighboridandadt:
                    neighborrank2neighboridandadt[neighbor_rank] = []
                neighborrank2neighboridandadt[neighbor_rank].append(
                    (neighbor_id, neighbor_adt)
                )

        # Estimate the size of a single value in neighborrank2neighboridandadt
        if neighborrank2neighboridandadt:
            sample_value = next(iter(neighborrank2neighboridandadt.values()))[0]
            estimated_value_size = sys.getsizeof(
                sample_value
            )  # Approximate size in bytes
            chunk_size = max(
                1, 1024 // estimated_value_size
            )  # Ensure at least one value per chunk
        else:
            chunk_size = 1
        # Send chunk nums
        sends_num_chunks_requests = []
        torank2numchunks = {}
        other_ranks = [(worker + i) % num_workers for i in range(1, num_workers)]
        for to_rank in other_ranks:
            if to_rank in neighborrank2neighboridandadt:
                # Send the data for this rank
                data_to_send_to_rank = neighborrank2neighboridandadt[to_rank]
                # Break the data into chunks
                num_chunks = len(data_to_send_to_rank) // chunk_size + (
                    1 if len(data_to_send_to_rank) % chunk_size > 0 else 0
                )
                torank2numchunks[to_rank] = num_chunks
                sends_num_chunks_requests.append(
                    comm.isend(
                        num_chunks,
                        dest=to_rank,
                        tag=0,
                    )
                )
        # Receive num_chunks from all ranks
        recvs_num_chunks_requests = []
        for from_rank in other_ranks:
            recvs_num_chunks_requests.append(comm.irecv(source=from_rank, tag=0))
        MPI.Request.waitall(sends_num_chunks_requests)
        recv_chunk_nums = MPI.Request.waitall(recvs_num_chunks_requests)

        # Send the chunks
        send_chunk_requests = []
        for to_rank in other_ranks:
            if to_rank in neighborrank2neighboridandadt:
                # Send the data for this rank
                data_to_send_to_rank = neighborrank2neighboridandadt[to_rank]
                num_chunks = torank2numchunks[to_rank]
                for i in range(num_chunks):
                    chunk = data_to_send_to_rank[i * chunk_size : (i + 1) * chunk_size]
                    send_chunk_request = comm.isend(
                        chunk,
                        dest=to_rank,
                        tag=i + 1,
                    )
                    if i >= len(send_chunk_requests):
                        send_chunk_requests.append([])
                    send_chunk_requests[i].append(send_chunk_request)
        # Receive the chunks
        recv_chunk_requests = []
        for i, from_rank in enumerate(other_ranks):
            num_chunks = recv_chunk_nums[i]
            for j in range(num_chunks):
                received_chunk_request = comm.irecv(source=from_rank, tag=j + 1)
                if j >= len(recv_chunk_requests):
                    recv_chunk_requests.append([])
                recv_chunk_requests[j].append(received_chunk_request)

        received_data = []
        num_send_chunk_requests = len(send_chunk_requests)
        num_recv_chunk_requests = len(recv_chunk_requests)
        for i in range(max(num_send_chunk_requests, num_recv_chunk_requests)):
            if i < num_send_chunk_requests:
                MPI.Request.waitall(send_chunk_requests[i])
            if i < num_recv_chunk_requests:
                received_data_ranks_chunk = MPI.Request.waitall(recv_chunk_requests[i])
                for received_data_rank_chunk in received_data_ranks_chunk:
                    received_data.extend(received_data_rank_chunk)

        for agent_id, modified_adts in received_data:
            agent_idx = self._rank2agentid2agentidx[worker][agent_id]
            original_adts = [adt[agent_idx] for adt in agent_data_tensors]
            reduce_result = reduce_func(original_adts, modified_adts)
            for prop_idx in range(self.num_properties):
                agent_data_tensors[prop_idx][agent_idx] = reduce_result[prop_idx]



"""
This file describes the different spaces that SAGESim agents
may exist in. Space is important in calculating proximity.
Promixity is important in approximating interactions.
Intractions are important when binning agents into
workers, which is the crux of load balancing...

"""

from typing import List, Any


class Space:
    """
    Abstract class for space

    """

    def __init__(
        self, neighbor_compute_func, locations_max_dims, locations_defaults
    ) -> None:
        """
        self_agent_adj_list hold
        """
        self._locations: List[Any] = []
        self._neighbor_compute_func = neighbor_compute_func
        self._locations_max_dims = locations_max_dims
        self._locations_defaults = locations_defaults
        self._agent_factory = None

    def get_location(self, agent_id: int) -> set:
        """Returns agents location"""
        return self._locations[agent_id]

    def add_agent(self, agent: int) -> None:
        """Adds agent to space"""
        self._locations = None

    def get_neighbors(self, agent_id: int) -> set:
        """Returns agents neighbors"""
        return self._neighbor_compute_func(self._locations, agent_id)

    ###NOTE: we need a function for load balancing


def _network_space_compute_neighbors(agent_locations):
    agents_current_neighbors = agent_locations
    return agents_current_neighbors


class NetworkSpace(Space):
    """Defines a NetworkSpace"""

    def __init__(self) -> None:
        """
        Uses super()._neighbors to hold adj list of the network
        """
        locations_max_dims = [0]
        locations_defaults = []
        super().__init__(
            _network_space_compute_neighbors, locations_max_dims, locations_defaults
        )

    def add_agent(self, agent: int) -> None:
        self._locations.append(set())
        self._agent_factory.set_agent_property_value(
            "locations",
            agent,
            self._locations[agent],
        )

    def connect_agents(
        self, agent_0: int, agent_1: int, directed: bool = False
    ) -> None:
        agent_0 = int(agent_0)
        agent_1 = int(agent_1)
        self._locations[agent_0].add(agent_1)

        if not directed:
            self._locations[agent_1].add(agent_0)

    def disconnect_agents(
        self, agent_0: int, agent_1: int, directed: bool = False
    ) -> None:

        self._locations[agent_0].remove(agent_1)
        if not directed:
            self._locations[agent_1].remove(agent_0)

    ###NOTE: we need a function for load balancing


if __name__ == "__main__":
    ns = NetworkSpace()
    for i in range(6):
        ns.add_agent(i)

    for i in range(6):
        print(f"Neighbors of {i}: {ns.get_neighbors(i)}")

    for i in [1, 3, 4]:
        ns.connect_agents(0, i)

    for i in range(6):
        print(f"Neighbors of {i}: {ns.get_neighbors(i)}")



from random import random

import cupy as cp
from cupyx import jit
from sagesim.utils import (
    get_this_agent_data_from_tensor,
    set_this_agent_data_from_tensor,
    get_neighbor_data_from_tensor,
)


# Define the step function to be registered for SIRBreed
@jit.rawkernel(device="cuda")
def step_func(
    tick,
    agent_index,
    globals,
    agent_ids,
    breeds,
    locations,
    state_tensor,
    preventative_measures_tensor,
):
    """
    At each simulation step, this function evaluates a subset of agents—either all agents in a serial run or a partition assigned to
    a specific rank in parallel processing—and determines whether an agent's state should change based on interactions with its neighbors
    and their respective preventative behaviors.

    Parameters:
    ----------
    tick : int
        The current simulation tick, which is the first item in the globals array.
    agent_index : int
        Index of the agent being evaluated in the agent_ids list.
    globals : cupy.array
        Global parameters;
        the first item will be our infection probability.
        the second item will be our recovery probability.
    agent_ids : cupy.array
        The tensor that contains the IDs of all agents assigned to the current rank, and their neighbors.
    breeds : cupy.array
        1D cupy.array of breed objects (unused here as we only have one type of breed, but must passed for interface compatibility).
    locations : cupy.array
        Adjacency list specifying neighbors for each agent, padded with cp.nan.
    state_tensor : cupy.array
        1D cupy array of current state of each agent.
    preventative_measures_tensor : cupy.array[]
        2D cupy array of vectors representing each agent’s preventative behaviors.
    Returns:
    -------
    None
        The function updates the `states` list in-place if an agent becomes infected.
    """
    # Get the list of neighboring agent IDs for the current agent based on network topology
    neighbor_ids = locations[agent_index]

    # Draw a random float in [0, 1) for stochastic decision-making
    rand = (
        random()
    )  # can replace with step_func_helper_get_random_float(rng_states, id)

    # Retrieve the global infection and recovery probabilities defined in the model
    p_infection = globals[0]
    p_recovery = globals[1]

    # Get the current state of the agent (e.g., susceptible, infected, recovered)
    agent_state = int(get_this_agent_data_from_tensor(agent_index, state_tensor))

    # Get the preventative measures vector for the current agent
    agent_preventative_measures = get_this_agent_data_from_tensor(
        agent_index, preventative_measures_tensor
    )

    # If agent is infected and the recovery condition passes, update agent’s state
    if agent_state == 2 and rand < p_recovery:
        set_this_agent_data_from_tensor(
            agent_index, state_tensor, 3
        )  # Agent becomes infected
    elif agent_state == 1:
        # Loop through each neighbor ID
        i = 0
        while i < len(neighbor_ids) and not cp.isnan(neighbor_ids[i]):
            neighbor_id = neighbor_ids[i]

            # Retrieve the state of the neighbor (e.g., susceptible, infected, recovered)
            neighbor_state = int(
                get_neighbor_data_from_tensor(agent_ids, neighbor_id, state_tensor)
            )

            # Get the preventative measures vector of the neighbor
            neighbor_preventative_measures = get_neighbor_data_from_tensor(
                agent_ids, neighbor_id, preventative_measures_tensor
            )

            # Initialize cumulative safety score for the interaction
            abs_safety_of_interaction = 0.0

            # Calculate total safety of interaction based on pairwise product of measures
            for n in range(len(agent_preventative_measures)):
                for m in range(len(neighbor_preventative_measures)):
                    abs_safety_of_interaction += (
                        agent_preventative_measures[n]
                        * neighbor_preventative_measures[m]
                    )

            # Normalize the safety score to be in [0, 1]
            normalized_safety_of_interaction = abs_safety_of_interaction / (
                len(agent_preventative_measures) ** 2
            )

            # If neighbor is infected and the infection condition passes, update agent’s state
            if neighbor_state == 2 and rand < p_infection * (
                1 - normalized_safety_of_interaction
            ):
                set_this_agent_data_from_tensor(
                    agent_index, state_tensor, 2
                )  # Agent becomes infected
            i += 1



from sagesim.model import Model
from sagesim.space import NetworkSpace
from sfr_breed import SFRBreed


def reduce_agent_data_tensors_(adts_A, adts_B):
    result = []
    # breed would be same as first
    result.append(adts_A[0])
    # network would be same as first
    result.append(adts_A[1])
    # OSMnxid would be same as first
    result.append(adts_A[2])
    # Popularity would be samge as first
    result.append(adts_A[3])
    # Num of vehicles at this intersection would be summed
    result.append(adts_A[4] + adts_B[4])
    return result


class SFRModel(Model):

    def __init__(self) -> None:
        space = NetworkSpace()
        super().__init__(space)
        self._sfr_breed = SFRBreed()
        self.register_breed(breed=self._sfr_breed)
        self.register_reduce_function(reduce_agent_data_tensors_)

    def create_agent(self, popularity, vehicle_num):
        agent_id = self.create_agent_of_breed(
            self._sfr_breed,
            popularity=popularity,
            vehicle_num=vehicle_num,
        )
        self.get_space().add_agent(agent_id)
        return agent_id

    def connect_agents(self, agent_0, agent_1):
        self.get_space().connect_agents(agent_0, agent_1)



from typing import Any, Callable, List, Dict, Optional, Union
from collections import OrderedDict
from math import nan


class Breed:
    def __init__(self, name: str) -> None:
        # self._properties is a dict with keys as property name and
        #   values are properties.
        #   properties themselves are list of type and default value.
        self._properties: OrderedDict[str, List[Any, Any]] = OrderedDict()
        self._prop2pos: Dict[str, int] = {}
        self._name: str = name
        self._step_funcs: Dict[int, Callable] = {}
        self._breedidx: int = -1
        self._num_properties: int = 0
        self._prop2maxdims: Dict[str, List[int]] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def properties(self) -> Dict[str, Any]:
        return self._properties

    @property
    def step_funcs(self) -> Dict[int, Callable]:
        return self._step_funcs

    def register_property(
        self,
        name: str,
        default: Union[int, float, List] = nan,
        max_dims: Optional[List[int]] = None,
    ) -> None:
        self._properties[name] = default
        self._prop2pos[name] = self._num_properties
        self._num_properties += 1
        self._prop2maxdims[name] = max_dims

    def register_step_func(
        self, step_func: Callable, module_fpath: str, priority: int = 0
    ):
        """
        What the agent is supposed to do during a simulation step.

        """
        self._step_funcs[priority] = (step_func, module_fpath)



from typing import List, Any

import awkward as ak
import numpy as np
import cupy as cp


def convert_to_equal_side_tensor(ragged_list: List[Any]) -> cp.array:
    awkward_array = ak.from_iter(ragged_list)
    assert (
        awkward_array.layout.minmax_depth[0] == awkward_array.layout.minmax_depth[1]
    ), "Tensor is of unequal depth"

    i = 1
    while i < awkward_array.layout.minmax_depth[0]:
        awkward_array = ak.fill_none(awkward_array, [], axis=i - 1)
        nums_in_level = ak.fill_none(ak.ravel(ak.num(awkward_array, axis=i)), value=0)
        awkward_array = ak.pad_none(
            awkward_array, int(max(nums_in_level)), axis=i, clip=True
        )
        i += 1

    awkward_array = ak.fill_none(awkward_array, np.nan, axis=-1)

    return ak.to_cupy(awkward_array).astype(np.float32)


def compress_tensor(regular_tensor: cp.array, min_axis: int = 1) -> List[Any]:
    awkward_tensor = ak.from_cupy(regular_tensor)
    awkward_tensor = ak.nan_to_none(awkward_tensor)
    awkward_tensor = ak.drop_none(awkward_tensor)

    i = -1
    while awkward_tensor.layout.minmax_depth[0] + i > min_axis:
        awkward_tensor = ak.mask(awkward_tensor, ak.num(awkward_tensor, axis=i) > 0)
        awkward_tensor = ak.drop_none(awkward_tensor)
        i -= 1

    return ak.to_list(awkward_tensor)



from cupyx import jit
import os
import sys
module_path = os.path.abspath('/home/co1/sagesim_github/SAGESim/examples/sir')
if module_path not in sys.path:
	sys.path.append(module_path)
from sir_step_func import *


@jit.rawkernel(device='cuda')
def stepfunc(
global_tick,
device_global_data_vector,
a0,a1,a2,a3,
sync_workers_every_n_ticks,
num_rank_local_agents,
agent_ids,
):
	thread_id = jit.blockIdx.x * jit.blockDim.x + jit.threadIdx.x
	agent_index = thread_id
	if agent_index < num_rank_local_agents:
		breed_id = a0[agent_index]
		for tick in range(sync_workers_every_n_ticks):

			thread_local_tick = int(global_tick) + tick

			if breed_id == 0:
				step_func(
					thread_local_tick,
					agent_index,
					device_global_data_vector,
					agent_ids,
					a0,a1,a2,a3,
				)


import unittest
import sagesim
import sagesim.breed

class TestSAGESimBasic(unittest.TestCase):
    def test_version(self):
        # Check if sagesim has a __version__ attribute
        self.assertTrue(hasattr(sagesim, '__version__'))

    def test_basics(self):
        model = sagesim.Model()
        breed_predator = sagesim.breed.Breed('predator')
        breed_predator
        breed_prey = sagesim.breed.Breed('prey')

        model.register_breed(breed_predator)
        model.register_breed(breed_prey)


        

if __name__ == '__main__':
    unittest.main()


from time import time
from random import random, sample

import networkx as nx
from mpi4py import MPI

from sir_model import SIRModel
from state import SIRState

if __name__ == "__main__":

    num_agents = 1000
    num_init_connections = 2
    rewiring_prob = 0.1

    num_infected = 10

    # Generate the Contact Network
    network = nx.watts_strogatz_graph(
        num_agents, num_init_connections, rewiring_prob, seed=42
    )

    # Instantiate the SIR Model
    model = SIRModel(p_infection=0.6, p_recovery=0.00)

    # Create agents
    for n in network.nodes:
        preventative_measures = [random() for _ in range(100)]
        model.create_agent(SIRState.SUSCEPTIBLE.value, preventative_measures)

    # Connect agents in the network
    for edge in network.edges:
        model.connect_agents(edge[0], edge[1])

    # Infect a random sample of agents
    for n in sample(sorted(network.nodes), num_infected):
        model.set_agent_property_value(n, "state", SIRState.INFECTED.value)

    model.setup(use_gpu=True)  # Enables GPU acceleration if available

    # # MPI environment setup
    comm = MPI.COMM_WORLD
    num_workers = comm.Get_size()
    worker = comm.Get_rank()

    # Run the simulation with 1 rank, and measure the time taken
    simulate_start = time()
    model.simulate(ticks=10, sync_workers_every_n_ticks=1)
    simulate_end = time()
    simulate_duration = simulate_end - simulate_start

    result = [
        SIRState(model.get_agent_property_value(agent_id, property_name="state"))
        for agent_id in range(num_agents)
        if model.get_agent_property_value(agent_id, property_name="state") is not None
    ]

    # count the number of infected agents
    num_infected = sum(1 for state in result if state == SIRState.INFECTED)
    num_recovered = sum(1 for state in result if state == SIRState.RECOVERED)
    num_susceptible = sum(1 for state in result if state == SIRState.SUSCEPTIBLE)

    if worker == 0:
        print(f"Simulation took {simulate_duration:.2f} seconds.")
        print(f"Number of infected agents: {num_infected}")
        print(f"Number of recovered agents: {num_recovered}")
        print(f"Number of susceptible agents: {num_susceptible}")



"""
SuperNeuroABM basic Model class

"""

from typing import Dict, List, Callable, Set, Any, Union
import os
from pathlib import Path
import importlib
import pickle
import math
import heapq
import warnings

import cupy as cp
import numpy as np
from mpi4py import MPI

from sagesim.agent import AgentFactory, Breed
from sagesim.space import Space
from sagesim.internal_utils import convert_to_equal_side_tensor

comm = MPI.COMM_WORLD
num_workers = comm.Get_size()
worker = comm.Get_rank()


class Model:

    def __init__(
        self,
        space: Space,
        threads_per_block: int = 32,
        step_function_file_path: str = "step_func_code.py",
    ) -> None:
        self._threads_per_block = threads_per_block
        self._step_function_file_path = step_function_file_path
        self._agent_factory = AgentFactory(space)
        self._is_setup = False
        self.globals = {}
        self.tick = 0
        # following may be set later in setup if distributed execution

    def register_breed(self, breed: Breed) -> None:
        if self._agent_factory.num_agents > 0:
            raise Exception(f"All breeds must be registered before agents are created!")
        self._agent_factory.register_breed(breed)

    def create_agent_of_breed(self, breed: Breed, add_to_space=True, **kwargs) -> int:
        agent_id = self._agent_factory.create_agent(breed, **kwargs)
        if add_to_space:
            self.get_space().add_agent(agent_id)
        return agent_id

    def get_agent_property_value(self, id: int, property_name: str) -> Any:
        if self._is_setup:
            self._agent_factory._update_agent_property(
                self.__rank_local_agent_data_tensors, id, property_name
            )
        return self._agent_factory.get_agent_property_value(
            property_name=property_name, agent_id=id
        )

    def set_agent_property_value(self, id: int, property_name: str, value: Any) -> None:
        self._agent_factory.set_agent_property_value(
            property_name=property_name, agent_id=id, value=value
        )

    def get_space(self) -> Space:
        return self._agent_factory._space

    def get_agents_with(self, query: Callable) -> Set[List[Any]]:
        return self._agent_factory.get_agents_with(query=query)

    def register_global_property(
        self, property_name: str, value: Union[float, int]
    ) -> None:
        self.globals[property_name] = value

    def set_global_property_value(
        self, property_name: str, value: Union[float, int]
    ) -> None:
        self.globals[property_name] = value

    def get_global_property_value(self, property_name: str) -> Union[float, int]:
        return self.globals[property_name]

    def register_reduce_function(self, reduce_func: Callable) -> None:
        self._reduce_func = reduce_func

    def setup(self, use_gpu: bool = True) -> None:
        """
        Must be called before first simulate call.
        Initializes model and resets ticks. Readies step functions
        and for breeds.

        :param use_cuda: runs model in GPU mode.
        :param num_dask_worker: number of dask workers
        :param scheduler_fpath: specify if using external dask cluster. Else
            distributed.LocalCluster is set up.
        """
        self._use_gpu = use_gpu
        # Create record of agent step functions by breed and priority
        self._breed_idx_2_step_func_by_priority: List[Dict[int, Callable]] = []
        heap_priority_breedidx_func = []
        for breed in self._agent_factory.breeds:
            for priority, func in breed.step_funcs.items():
                heap_priority_breedidx_func.append((priority, (breed._breedidx, func)))
        heapq.heapify(heap_priority_breedidx_func)
        last_priority = None
        while heap_priority_breedidx_func:
            priority, breed_idx_func = heapq.heappop(heap_priority_breedidx_func)
            if last_priority == priority:
                # same slot in self._breed_idx_2_step_func_by_priority
                self._breed_idx_2_step_func_by_priority[-1].update(
                    {breed_idx_func[0]: breed_idx_func[1]}
                )
            else:
                # new slot
                self._breed_idx_2_step_func_by_priority.append(
                    {breed_idx_func[0]: breed_idx_func[1]}
                )
                last_priority = priority

        # Generate global data tensor
        self._global_data_vector = list(self.globals.values())
        if worker == 0:
            with open(self._step_function_file_path, "w") as f:
                f.write(
                    generate_gpu_func(
                        self._agent_factory.num_properties,
                        self._breed_idx_2_step_func_by_priority,
                    )
                )
        comm.barrier()
        # Generate agent data tensors
        self.__rank_local_agent_data_tensors = (
            self._agent_factory._generate_agent_data_tensors()
        )
        self._is_setup = True

    def simulate(
        self,
        ticks: int,
        sync_workers_every_n_ticks: int = 1,
    ) -> None:
        comm.barrier()
        # Import the package using module package
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            step_func_module = importlib.import_module(
                os.path.splitext(self._step_function_file_path)[0]
            )

        # Access the step function using the module
        self._step_func = step_func_module.stepfunc

        # Repeatedly execute worker coroutine until simulation
        # has run for the right amount of ticks
        original_sync_workers_every_n_ticks = sync_workers_every_n_ticks
        for time_chunk in range((ticks // original_sync_workers_every_n_ticks) + 1):

            if time_chunk == (ticks // original_sync_workers_every_n_ticks):
                # Final chunk: handle remaining ticks
                remaining_ticks = ticks - (
                    time_chunk * original_sync_workers_every_n_ticks
                )
                if remaining_ticks == 0:
                    break
                sync_workers_every_n_ticks = remaining_ticks
            else:
                # Regular chunk: use original batch size
                sync_workers_every_n_ticks = original_sync_workers_every_n_ticks

            self.worker_coroutine(sync_workers_every_n_ticks)

    def save(self, app: "Model", fpath: str) -> None:
        """
        Saves model. Must be overridden if additional data
        pertaining to application must be saved.

        :param fpath: file path to save pickle file at
        :param app_data: additional application data to be saved.
        """
        if "_agent_data_tensors" in app.__dict__:
            del app.__dict__["_agent_data_tensors"]
        with open(fpath, "wb") as fout:
            pickle.dump(app, fout)

    def load(self, fpath: str) -> "Model":
        """
        Loads model from pickle file.

        :param fpath: file path to pickle file.
        """
        with open(fpath, "rb") as fin:
            app = pickle.load(fin)
        return app

    # Define worker coroutine that executes cuda kernel
    # ------------------------------------------------------
    def worker_coroutine(
        self,
        sync_workers_every_n_ticks,
    ):
        """
        Corountine that exec's cuda kernel. This coroutine should
        eventually be distributed among dask workers with agent
        data partitioning and data reduction.

        :param device_global_data_vector: cuda device array containing
            SAGESim global properties
        :param agent_data_tensors: listof property data tensors defined by user.
            Contains all agent info. Each inner list represents a particular property and may
            itself be a multidimensional list. This is also where the
            cuda kernels will make modifications as agent properties
            are updated.
        :param current_tick: Current simulation tick
        :param sync_workers_every_n_ticks: number of ticks to forward
            the simulation by
        :param agent_ids: agents to process by this cudakernel call
        """

        self.__rank_local_agent_ids = list(
            self._agent_factory._rank2agentid2agentidx[worker].keys()
        )
        threadsperblock = 32
        blockspergrid = int(
            math.ceil(len(self.__rank_local_agent_ids) / threadsperblock)
        )
        rank_local_agents_neighbors = self.get_space()._neighbor_compute_func(
            self.__rank_local_agent_data_tensors[1]
        )
        (
            self.__rank_local_agent_ids,
            self.__rank_local_agent_data_tensors,
            received_neighbor_ids,
            received_neighbor_adts,
        ) = self._agent_factory.contextualize_agent_data_tensors(
            self.__rank_local_agent_data_tensors,
            self.__rank_local_agent_ids,
            rank_local_agents_neighbors,
        )
        rank_local_agent_and_neighbor_adts = [
            convert_to_equal_side_tensor(
                self.__rank_local_agent_data_tensors[i] + received_neighbor_adts[i]
            )
            for i in range(self._agent_factory.num_properties)
        ]
        self._global_data_vector = cp.array(self._global_data_vector)
        rank_local_agent_and_non_local_neighbor_ids = cp.array(
            self.__rank_local_agent_ids + received_neighbor_ids
        )
        self._step_func[blockspergrid, threadsperblock](
            self.tick,
            self._global_data_vector,
            *rank_local_agent_and_neighbor_adts,
            sync_workers_every_n_ticks,
            cp.float32(len(self.__rank_local_agent_ids)),
            rank_local_agent_and_non_local_neighbor_ids,
        )
        # Update global tick counter after all threads have completed
        self.tick += sync_workers_every_n_ticks
        cp.get_default_memory_pool().free_all_blocks()
        num_agents = len(self.__rank_local_agent_ids)
        self.__rank_local_agent_data_tensors = [
            rank_local_agent_and_neighbor_adts[i][:num_agents].tolist()
            for i in range(self._agent_factory.num_properties)
        ]
        """worker_agent_and_neighbor_data_tensors = (
            self._agent_factory.reduce_agent_data_tensors(
                worker_agent_and_neighbor_data_tensors,
                agent_and_neighbor_ids_in_subcontext,
                self._reduce_func,
            )
        )
        """
        self._global_data_vector = comm.allreduce(
            self._global_data_vector.tolist(), op=reduce_global_data_vector
        )


def reduce_global_data_vector(A, B):
    values = np.stack([A, B], axis=1)
    return np.max(values, axis=1)


def generate_gpu_func(
    n_properties: int,
    breed_idx_2_step_func_by_priority: List[List[Union[int, Callable]]],
) -> str:
    """
    cupy jit.rawkernel does not like us passing *args into
    them. This is because the Python function
    will be compiled by cupy.jit and the parameter arguments
    type and count must be set at jit compilation time.
    However, SAGESim users will have varying numbers of
    properties in their step functions, which means
    our cuda kernel's parameter count would also be variable.
    Normally, we'd just define the stepfunc with *args, but
    due to the above constraints we have to infer the number of
    arguments from the user defined breed step functions,
    rewrite the overall stepfunc as a string and then pass it
    into cupy.jit to be compiled.

    This function returns a str representation of stepfunc cupy jit.rawkernel:

        step_funcs_code = generate_gpu_func(
                    len(agent_data_tensors),
                    breed_idx_2_step_func_by_priority,
                )
    This function can then be directly loaded using importlib or written to a
    file and imported. For example, if you write the code to a file
    called step_func_code.py, you can import it as below:

        import importlib
        step_func_module = importlib.import_module("step_func_code")
        stepfunc = step_func_module.stepfunc
    Then you can run the stepfunc as a jit.rawkernel as below:

        stepfunc[blockspergrid, threadsperblock](
                device_global_data_vector,
                *agent_data_tensors,
                current_tick,
                sync_workers_every_n_ticks,
            )
        )

    :param n_properties: int total number of agent properties
    :param breed_idx_2_step_func_by_priority: List of List. Each inner List
        first element is the breedidx and second element is a tuple of the user defined
        step function, and the file where it is defined.
        The major list elements are ordered in decreasing order of execution
        priority
    :return: str representation of stepfunc cuda kernal
        that can be written to file or imported directly.

    """
    args = [f"a{i}" for i in range(n_properties)]
    sim_loop = []
    step_sources = ["import os", "import sys"]
    imported_modules = set()
    for breed_idx_2_step_func in breed_idx_2_step_func_by_priority:
        for breedidx, breed_step_func_info in breed_idx_2_step_func.items():
            breed_step_func_impl, module_fpath = breed_step_func_info
            step_func_name = getattr(breed_step_func_impl, "__name__", repr(callable))
            module_fpath = Path(module_fpath).absolute()
            module_name = module_fpath.stem
            if module_fpath not in imported_modules:
                step_sources += [
                    f"module_path = os.path.abspath('{module_fpath.parent}')",
                    "if module_path not in sys.path:",
                    "\tsys.path.append(module_path)",
                    f"from {module_name} import *",
                ]
                imported_modules.add(module_fpath)
            sim_loop += [
                f"if breed_id == {breedidx}:",
                f"\t{step_func_name}(",
                "\t\tthread_local_tick,",
                "\t\tagent_index,",
                "\t\tdevice_global_data_vector,",
                "\t\tagent_ids,",
                f"\t\t{','.join(args)},",
                "\t)",
            ]
    step_sources = "\n".join(step_sources)

    # Preprocess parts that would break in f-strings
    joined_sim_loop = "\n\t\t\t".join(sim_loop)
    joined_args = ",".join(args)

    func = [
        "from cupyx import jit",
        step_sources,
        "\n\n@jit.rawkernel(device='cuda')",
        "def stepfunc(",
        "global_tick,",
        "device_global_data_vector,",
        joined_args + ",",
        "sync_workers_every_n_ticks,",
        "num_rank_local_agents,",
        "agent_ids,",
        "):",
        "\tthread_id = jit.blockIdx.x * jit.blockDim.x + jit.threadIdx.x",
        "\tagent_index = thread_id",
        "\tif agent_index < num_rank_local_agents:",
        "\t\tbreed_id = a0[agent_index]",
        "\t\tfor tick in range(sync_workers_every_n_ticks):",
        f"\n\t\t\tthread_local_tick = int(global_tick) + tick",
        f"\n\t\t\t{joined_sim_loop}",
    ]

    func = "\n".join(func)
    return func



from sagesim.breed import Breed
import cupy as cp
from cupyx import jit
from pathlib import Path


class SFRBreed(Breed):

    def __init__(self) -> None:
        name = "SFR"
        super().__init__(name)

        # register osmnxid
        # self.register_property("osmnxid")

        # popularity of the agent, real value between [0,1]
        self.register_property("popularity", default=0.5)

        # number of vehicles at the current agent/node/intersection
        self.register_property("vehicle_num", default=10)

        # register step function
        curr_fpath = Path(__file__).resolve()
        self.register_step_func(step_func, curr_fpath)


# Define the step function to be registered for SIRBreed
@jit.rawkernel(device="cuda")
def step_func(
    tick,
    agent_index,
    globals,
    agent_ids,
    breeds,
    locations,
    popularities,
    vehicle_nums,
):

    # Get agent's vehicle number and neighbors' locations
    agent_vehicle_num = vehicle_nums[agent_index]
    # Zero out agent's vehicle count
    vehicle_nums[agent_index] = agent_vehicle_num
    neighbor_ids = locations[agent_index]

    # find total popularity of all neighbors
    total_popularity = cp.float64(0)
    neighbor_i = 0
    while not cp.isnan(neighbor_ids[neighbor_i]) and neighbor_i < len(neighbor_ids):
        neighbor_id = neighbor_ids[neighbor_i]
        # Find the index of the neighbor_id in agent_ids
        neighbor_index = -1
        i = 0
        while i < len(agent_ids) and agent_ids[i] != neighbor_id:
            i += 1
        if i < len(agent_ids):
            neighbor_index = i
            neighbor_popularity = popularities[int(neighbor_index)]
            total_popularity += neighbor_popularity
        neighbor_i += 1

    remainder = agent_vehicle_num
    largest_alloc = 0

    if total_popularity > 0:
        remainder_alloc_index = -1
        neighbor_i = 0
        while not cp.isnan(neighbor_ids[neighbor_i]) and neighbor_i < len(neighbor_ids):
            neighbor_id = neighbor_ids[neighbor_i]
            # Find the index of the neighbor_id in agent_ids
            neighbor_index = 0
            while (
                neighbor_index < len(agent_ids)
                and agent_ids[neighbor_index] != neighbor_id
            ):
                neighbor_index += 1
            if (neighbor_index < len(agent_ids)) and (agent_ids[i] != neighbor_id):
                neighbor_popularity = popularities[int(neighbor_index)]

                neighbor_allocation = int(
                    agent_vehicle_num * neighbor_popularity / total_popularity
                )
                # find the top popularity neighbor
                if neighbor_allocation > largest_alloc:
                    remainder_alloc_index = neighbor_index
                    largest_alloc = neighbor_allocation

                remainder -= neighbor_allocation
            neighbor_i += 1

        # Distribute the remainder (due to rounding) to top contributors
        if remainder > 0 and remainder_alloc_index >= 0:
            vehicle_nums[int(remainder_alloc_index)] += remainder
